/***{{{*******************************************************************
 * This file is part of libco - object library for C                     *
 * Copyright (c) 2002                                                    *
 *     Peter Rockai (yenar) <yenar@host.sk>                              *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2 of the License, or (at your option) any later version.      *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330,                         *
 * Boston, MA 02111-1307  USA                                            *
 *******************************************************************}}}***/
/* {{{ file description */
/**
 * @file co_exception.c
 * @brief Cleanup stack exception handling implementation.
 **/
/* }}} */

/* {{{ includes */
#include <stdlib.h>
#include <stdio.h>
#include "co_exception.h"
#include "co_debug.h"
#include "co_util.h"
/* }}} */

#include "co_exception.cometa"

//static cox_state *cox_state_simple (void);
static cox_state *cox_state_new (void);

#if CO_THREADS

// FIXME!
#include <pthreads.h>
#define STATE (__cox_state ())
cox_state *__cox_state (void)
{
}

#else // ! CO_THREADS

static cox_state *__cox_state;
#define STATE (__cox_state)

#endif /* CO_THREADS */

/* {{{ */ int *_cox_get_mark (cox_state *state)
	/// Get current stack top position.
{
	int i = 0;
	int *marks = malloc ((state -> nstacks + 1) * sizeof (int));
	if (!marks)
		return 0;
	marks [0] = state -> nstacks;
	for (; i < state -> nstacks; i ++) {
		marks [i + 1] = state -> stacks [i] ->stack_top;
	}
	return marks;
} /* }}} */
/* {{{ */
int *cox_get_mark (void)
{
	return _cox_get_mark (STATE);
}
/* }}} */

/* {{{ */
unsigned _cox_handler_pop (cox_handler *_h)
{
	cox_handler *h;
	h = STATE -> cur_h;
	CO_DEBUG (20, "popping handler: %p (mark [0] = %d)", h,
			h ? h -> stack_mark [0] : -1);
	assert (h == _h);
	STATE -> cur_h = h -> n;
	free (_h -> stack_mark);
	return _h -> error;
}
/* }}} */
/* {{{ */
void _cox_handler_push (cox_handler *h)
{
	h -> n = __cox_state -> cur_h;
	h -> stack_mark = cox_get_mark ();
	h -> error = ERR_OK;
	__cox_state -> cur_h = h;
	CO_DEBUG (18, "pushed handler: %p (mark = %d)", h, h -> stack_mark);
}
/* }}} */

/* {{{ */
void _cox_stack_grow (cox_stack *st)
{
	cox_stack_seg *seg;
	CO_DEBUG (8, "growing cleanup stack");
	seg = malloc (sizeof (cox_stack_seg));
	if (!seg)
		COX_THROW (nomem);
	seg -> n = st -> top;
	st -> top = seg;
	st -> stack_size += _COX_STACK_SIZE;
}
/* }}} */
/* {{{ */
void _cox_stack_shrink (cox_stack *st)
	/// Free unused memory from cleanup stack.
{
	cox_stack_seg *seg;
	CO_DEBUG (5, "shrinking cleanup stack %p", st);
	seg = st -> top;
	st -> top = seg -> n;
	free (seg);
	st -> stack_size -= _COX_STACK_SIZE;
}
/* }}} */

/* {{{ */
void _cox_stack_unwind (cox_stack *st, int mark)
	/// Remove all items off top of stack up to @a mark.
{
	unsigned idx;
	cox_stack_it *item;
	int i = st -> stack_top - mark;
	assert (mark >= 0);
	/* CO_DEBUG (8, "topmost handler: %p (mark = %d)", __cox_state -> cur_h,
			__cox_state -> cur_h ?
			__cox_state -> cur_h -> stack_mark : -1); */
	CO_DEBUG (8, "unwinding last %d item(s)", i);
	assert (i >= 0); // FIXME?
	for (; i > 0; i --) {
		assert (st -> stack_top > 0);
		st -> stack_top --;
		idx = st -> stack_top % _COX_STACK_SIZE;
		item = &(st -> top -> items [idx]);
		CO_DEBUG (1, "unwinding: item %p with idx %d", item -> item, idx);
		item -> cleanup (item -> item, item -> cleanup_data);
		if (! idx && st -> stack_size > _COX_STACK_SIZE)
			_cox_stack_shrink (st);
	}
}
/* }}} */
/* {{{ */
void cox_stack_unwind (int stack, int mark)
{
	assert (stack < STATE -> nstacks && stack >= 0);
	_cox_stack_unwind (STATE -> stacks [stack], mark);
}
/* }}} */
/* {{{ */
void cox_unwind (int *marks)
{
	int i;
	assert ((*marks) == STATE -> nstacks);
	for (i = 0; i < *marks; i ++) {
		cox_stack_unwind (i, marks [i + 1]);
	}
}
/* }}} */

/* {{{ */
void _cox_stack_push (cox_stack *stack, void *_item,
		cox_cleanup_f cleanup, const void *data)
{
	if (stack -> stack_size == stack -> stack_top) {
		_cox_stack_grow (stack);
	}
	{
		unsigned idx = (stack -> stack_top % _COX_STACK_SIZE);
		cox_stack_it *item = &(stack -> top -> items [idx]);
		CO_DEBUG (8, "pushing item %p with idx %d", _item, idx);
		item -> item = _item;
		item -> cleanup_data = data;
		item -> cleanup = cleanup;
		stack -> stack_top ++;
		CO_DEBUG (32, "complete");
	}
} /* }}} */
/* {{{ */
void _cox_stack_pop (cox_stack *st, void *_item, bool clean)
	/// Pop cleanup handler and function, possibly free it.
	/**
	 * @param _item - pointer to item being popped
	 * @param clean - if true, we will clean the item, otherwise it will
	 *        remain intact
	 **/
{
	unsigned idx;
	cox_stack_it *item;

	st -> stack_top --;
	idx = st -> stack_top % _COX_STACK_SIZE;
	item = &(st -> top -> items [idx]);

	CO_DEBUG (8, "popping item %p (topmost item = %p, idx %d)", _item,
			item -> item, idx);

	/* if (item -> item != _item) {
		_cox_state_unwind (0);
		assert (0); // ehm ehm
	} */
	assert (item -> item == _item);

	if (clean)
		item -> cleanup (item -> item, item -> cleanup_data);
	/*if (__cox_state -> stack_size > _COX_STACK_SIZE
			&& (__cox_state -> stack_size - __cox_state -> stack_top)
			    >= (_COX_STACK_SIZE)) {
		cox_stack_shrink ();
	}*/
	if (! idx && st -> stack_size > _COX_STACK_SIZE)
		_cox_stack_shrink (st);
} /* }}} */
/* {{{ */
void cox_stack_pop (int stack, void *_item, bool clean)
{
	assert (stack < STATE -> nstacks && stack >= 0);
	CO_DEBUG (2, "popping from %d", stack);
	_cox_stack_pop (STATE -> stacks [stack], _item, clean);
}
/* }}} */
/* {{{ */
void cox_stack_push (int stack, void *_item,
		cox_cleanup_f cleanup, const void *data)
{
	assert (stack < STATE -> nstacks && stack >= 0);
	CO_DEBUG (2, "pushing to %d", stack);
	_cox_stack_push (STATE -> stacks [stack], _item, cleanup, data);
}
/* }}} */

/* {{{ */
cox_stack *_cox_stack_new (void)
	/// Create new cleanup stack.
{
	cox_stack *st;
	st = malloc (sizeof (cox_stack));
	if (! st)
		return 0;
	st -> top = malloc (sizeof (cox_stack_seg));
	if (! st -> top) {
		free (st);
		return 0;
	}
	st -> stack_size = _COX_STACK_SIZE;
	st -> stack_top = 0;
	return st;
}
/* }}} */
/* {{{ */
int cox_stack_create (void)
{
	int stackn = STATE -> nstacks;
	STATE -> nstacks ++;
	// FIXME: reliability
	STATE -> stacks = realloc (STATE -> stacks,
			(stackn + 1) * sizeof (cox_stack *));
	STATE -> stacks [stackn] = _cox_stack_new ();
	return stackn;
}
/* }}} */

/* {{{ */
cox_state *cox_state_new (void)
	/// Create new CSEH state object.
{
	cox_state *s;
	s = malloc (sizeof (cox_state));
	if (! s)
		return 0;
	s -> stacks = 0;
	s -> nstacks = 0;
	s -> cur_h = 0;
	return s;
}
/* }}} */
/* {{{ */
void cox_state_free (cox_state *s)
	/// Create new CSEH state object.
{
	int i;
	for (i = 0; i < s -> nstacks; i ++) {
		_cox_stack_unwind (s -> stacks [i], 0);
	}
	free (s);
}
/* }}} */
/* {{{ */
void __co_exception_init (void)
	/// Initialize CSEH subsystem.
{
	//cox_state *s;

	STATE = cox_state_new ();
	assert (STATE);
	cox_stack_create ();
	cox_stack_create ();
	//__cox_state_simple = s;
	//__cox_state = cox_state_simple;
}
/* }}} */
/* {{{ */
void __co_exception_fini (void)
	/// Initialize CSEH subsystem.
{
	//cox_unwind (0);
	cox_state_free (__cox_state);
}
/* }}} */

/* {{{ */
void __cox_throw (long long err)
	/// Throw exception.
{
	cox_handler *h;
	h = __cox_state -> cur_h;
	if (! h) return;
	h -> error = err;
	cox_unwind (h -> stack_mark);
	longjmp (h -> jb, h -> error);
}
/* }}} */


/* {{{ */
void *cox_malloc (size_t size)
	/// Allocate memory and push it onto cleanup stack.
	/**
	 * This also throws exception if program runs out of memory. Uses malloc()
	 * to get memory and free() to free it.
	 * @param size - size of block to be allocated.
	 **/
{
	void *p;
#if RANDOM_MEM_EXHAUSTION
	if (rand () < RAND_MAX / 500) COX_THROW (nomem);
#endif
	p = malloc (size);
	if (!p) {
		CO_DEBUG (3, "failed to allocate %u bytes of memory", size);
		COX_THROW (nomem);
	}
	cox_push (p, (cox_cleanup_f) free, 0);
	memset (p, 0, size);
	return p;
}
/* }}} */
/* {{{ co_*_new helper macros */
#define CO_ARRAY_ADM(last, ar, ar_s, type) \
	do { \
		if (! ar) { \
			ar_s = 30; \
			ar = new_ar_p (type, ar_s); \
		} \
		if (ar_s <= last) { \
			ar_s += 10; \
			ar = realloc (ar, sizeof (type) * ar_s); \
			memset (ar + ar_s - 10, 0, 10 * sizeof (type)); \
		} \
	} while (0)
/* }}} */

#if 0
/* {{{ */
cox_state *cox_state_simple (void)
	/// Get CSEH state object, function for non-threaded programs.
{
	return __cox_state_simple;
}
/* }}} */
/* {{{ */
const char *co_error_string (int id)
{
	return __exception_id_string [id];
}
/* }}} */
static const char **__exception_id_string = 0;
static int __exception_id_string_sz = 0;
/* {{{ */
int __co_exception_id_new (const char *descr)
{
	static int last = 1; // 0 is ERR_OK
	CO_DEBUG (2, "adding exception %d: %s", last, descr);
	CO_ARRAY_ADM (last, __exception_id_string,
			__exception_id_string_sz, char *);
	__exception_id_string [last] = descr;
	CO_DEBUG (2, "result: %s", co_error_string (last));
	++ last;
	return last - 1;
}
/* }}} */
/* {{{ */ static inline void *cox_realloc (void *p, size_t size)
	/// DO NOT USE!
{
	cox_pop (p);
	p = realloc (p, size);
	if (!p) COX_THROW (nomem);
	cox_push (p, (cox_cleanup_f)free, 0);
	return p;
} /* }}} */
#endif
