/***{{{*******************************************************************
 * This file is part of libco - object library for C                     *
 * Copyright (c) 2002, 2003                                              *
 *     Peter Rockai (yenar) <yenar@host.sk>                              *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2 of the License, or (at your option) any later version.      *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330,                         *
 * Boston, MA 02111-1307  USA                                            *
 *******************************************************************}}}***/
/* {{{ file description */
/**
 * @file co_object.h
 * @brief Object definition and function prototypes.
 **/
/* }}} */

#ifndef __CO_OBJECT_H__
#define __CO_OBJECT_H__

#include <stdarg.h>
#include "co_type.h"
#include "co_exception.h"
#include "co_debug.h"

/* {{{ co message macros */
/// Send message to object @a o (message selector is 2nd arg).
#define CO_M(o, ...) (CO_DEBUG (8, "sending msg..."), \
		co_m (o, o, COM (__VA_ARGS__)))
/// Send message to object @a o (message selector is 2nd arg).
#define CO_M_V(o, a, va) (CO_DEBUG (8, "sending msg..."), \
		co_m_v (o, COM (a), va))
/// Delegate message to @a o.
#define CO_D(o, ...) (co_m (o, this, COM (__VA_ARGS__)))
/// Delegate message to @a o.
#define CO_D_V(o, s, m, v) (co_m_v (o, s, m, v))
/// Get (set) object attribute of @a o (this is lvalue).
#define CO_A(t, o, ...) (CO_DEBUG (8, "getting attribute..."), \
		* (co_m (o, o, COM (__VA_ARGS__)) . _ ## t ## _p))
/* }}} */
/* {{{ co id construction macros */
/// Construct real co action identifier.
#define COM(...) __co_id_message_ ## __VA_ARGS__
/// Construct real co type identifier.
#define COT(...) __co_id_type_ ## __VA_ARGS__
/// Construct real co event identifier.
#define COE(...) __co_id_event_ ## __VA_ARGS__
/// Construct real co property identifier.
#define COP(...) __co_id_property_ ## __VA_ARGS__
/* }}} */

#define CO_PUSH(o) (cox_push ((o), (cox_cleanup_f) co_o_cleanup, 0))
#define CO_COPP_FLAGS(...)

/* {{{ CO_POSSIBLY_UNUSED */
#ifdef __GNUC__
#define CO_POSSIBLY_UNUSED __attribute__ ((unused))
#else
#define CO_POSSIBLY_UNUSED
#endif
/* }}} */

#define CO_OBJECT(x, f) struct x; CO_TYPEDEF (x); struct x
#define CO_ATTR(x, y) x y;

typedef unsigned co_id_t;

typedef co (*co_constructor) ();

/// Function pointer to object message handler.
typedef co_v (*co_message_handler) (co o, co s, co_id_t a, va_list ap);

/* {{{ */
#define CO_HEAD \
	co_id_t type; \
	co_message_handler msg_handle; \
	void *data
/* }}} */
/* {{{ */
struct co_o_minimal
{
	CO_HEAD;
};
CO_TYPEDEF (co_o_minimal);
/* }}} */

/// Send message @a a to object @a o, vararg version.
co_v co_m (co o, co s, co_id_t a, ...);

/// Send message to a object, va_list version.
co_v co_m_v (co o, co s, co_id_t a, va_list ap);

/// Delete object in a safe way.
/**
 * This function schedules object for destruction. This means: if object
 * still performs some actions, it will be destroyed when last one is
 * finished. Otherwise (the object is idle) it will be destroyed
 * immediately. This way, you can safely destroy objects from their
 * message handlers, without causing kind of lossage free()'ing them would
 * cause (You can't free them, because upon return from message handler
 * the state of object is manipulated - and it isn't wise to write into
 * free'd memory). This is equivalent to M (o, delete), for use with cleanup
 * stack.
 *
 * @param o - object to delete
 **/
void co_o_cleanup (co o);

#endif /* __CO_OBJECT_H__ */
