/***{{{*******************************************************************
 * This file is part of libco - object library for C                     *
 * Copyright (c) 2002                                                    *
 *     Peter Rockai (yenar) <yenar@host.sk>                              *
 *                                                                       *
 * This library is free software; you can redistribute it and/or         *
 * modify it under the terms of the GNU Lesser General Public            *
 * License as published by the Free Software Foundation; either          *
 * version 2 of the License, or (at your option) any later version.      *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 * Lesser General Public License for more details.                       *
 *                                                                       *
 * You should have received a copy of the GNU Lesser General Public      *
 * License along with this library; if not, write to the Free Software   *
 * Foundation, Inc., 59 Temple Place, Suite 330,                         *
 * Boston, MA 02111-1307  USA                                            *
 *******************************************************************}}}***/
/* {{{ file description */
/** @file co_super.c
 *  @brief Program super-object implementation. */
/* }}} */

/* {{{ includes */
#include "co_super.h"
#include "co_util.h"
#include "co_debug.h"
#include "co_conv.h"
#include "co_module.h"
/* }}} */

#include "co_super.cometa"

/* {{{ */
static int compar_find (const void *_key, const void *_memb)
{
	const char *key = _key;
	const co_id_head *memb = _memb;
	CO_DEBUG (2, "comparing: '%s' to '%s'", key, (*memb) -> name);
	return strcmp (key, (*memb) -> name);
}
/* }}} */
/* {{{ */
static int compar_sort (const void *_a, const void *_b)
{
	const co_id_head *a = _a;
	const co_id_head *b = _b;
	return strcmp ((* a) -> name, (* b) -> name);
}
/* }}} */
/* {{{ */
void *co_id_find (const co_id_array ar, const char *name)
{
	void **a = bsearch (name, ar -> shadow, ar -> n, sizeof (void *),
			compar_find);
	a = a ? *a : 0;
	CO_DEBUG (2, "found: %p", a);
	return a;
}
/* }}} */
/* {{{ */
static void co_id_sort (const co_id_array ar)
{
	qsort (ar -> shadow, ar -> n, sizeof (void *), compar_sort);
}
/* }}} */

#define __INIT_SIZE__ 32
/* {{{ */
static void *co_id_array_new (size_t memb_sz)
{
	co_id_array ar;
	ar = cox_malloc (sizeof (co_id_array_s));
	ar -> n = 0;
	ar -> memb_sz = memb_sz;
	ar -> size = __INIT_SIZE__;
	ar -> shadow = cox_malloc (sizeof (void *) * __INIT_SIZE__);
	ar -> master = cox_malloc (memb_sz * __INIT_SIZE__);
	cox_pop (ar -> master);
	cox_pop (ar -> shadow);
	cox_pop (ar);
	return ar;
}
/* }}} */
/* {{{ */
static void co_id_array_free (co_id_array ar)
{
	free (ar -> shadow);
	free (ar -> master);
	free (ar);
}
/* }}} */
/* {{{ */
static int co_id_add (co_id_array ar, void *it)
{
	if (ar -> n >= ar -> size) {
		int i;
		CO_DEBUG (2, "growing %p", ar);
		ar -> size *= 2;
		ar -> shadow = realloc (ar -> shadow, ar -> size * sizeof (void *));
		ar -> master = realloc (ar -> master, ar -> size * ar -> memb_sz);
		// reestablish proper pointers in shadow array
		// realloc might have moved master :-/
		for (i = 0; i < ar -> n; i ++)
			ar -> shadow [i] = ar -> master + i;
	}
	memcpy (ar -> master + ar -> n, it, ar -> memb_sz);
	ar -> shadow [ar -> n] = ar -> master + ar -> n;
	ar -> master [ar -> n] . id = ar -> n;
	CO_DEBUG (1, "adding item %p - #%03d - '%s'",
			ar, ar -> n, ar -> master [ar -> n] . name);
	ar -> n ++;
	co_id_sort (ar);
	return ar -> n - 1;
}
/* }}} */

/* {{{ */
CO_METH (co_super, module_load, const char *mod)
{
	CO_METH_INI (co_super, module_load);
	CO_RETURN (co, co_module_load (this, mod));
}
/* }}} */
/* {{{ */
CO_METH (co_super, exit, int exitcode)
{
	CO_METH_INI (co_super, exit);
	co_meth_array_free (((co_o_small) this) -> meth_ar);
	co_id_array_free (td -> id_type);
	co_id_array_free (td -> id_msg);
	co_id_array_free (td -> id_prop);
	co_id_array_free (td -> id_exception);
	co_id_array_free (td -> id_event);
	__co_module_fini ();
	cox_stack_pop (COX_STACK_MESG, this, false); // FIXME: depends on co_m_v impl
	free (td);
	free (this);
	__co_exception_fini ();
	exit (exitcode);
	CO_RETURN (int, 0);
}
/* }}} */

/* {{{ */
int co_id_init (co_id_array ar, const char *name)
{
	co_id_head m;
	CO_DEBUG (2, "getting %s", name);
	if ((m = co_id_find (ar, name))) {
		return m -> id;
	}
	m = alloca (ar -> memb_sz);
	m -> name = name;
	return co_id_add (ar, m);
}
/* }}} */

//void __co_mod_init (co_super);
void __co_mod_autoinit (co_super);

/* {{{ */
co co_init (void)
	/// Constructor for co_super.
{
	co_o_small this;
	co_super td;

	// initialize exception handling
	__co_exception_init ();

	this = new (co_super_s);
	td = new (co_super_s);
	co_o_small_bless (this);
	this -> data = td;
	cox_pop (td);

	td -> id_type = co_id_array_new (sizeof (co_id_head_s));
	td -> id_msg = co_id_array_new (sizeof (co_id_head_s));
	td -> id_prop = co_id_array_new (sizeof (co_id_head_s));
	td -> id_exception = co_id_array_new (sizeof (co_id_head_s));
	td -> id_event = co_id_array_new (sizeof (co_id_head_s));

	//Py_SetProgramName ("whumps");
	Py_Initialize ();

	__co_mod_autoinit (this);
	__co_module_init ();
	co_super_meth_ar_init (this);
	this -> type = COT (co_super);

	cox_pop (this);
	//CO_BLESS (this, co_super, 0);

	return this;
} /* }}} */
CO_METH (co_super, init)
{
	// XXX implement CO_COPP_FLAGS in copp
}

CO_COPP_FLAGS (
	no-constructor: co_super;
)
